#!/usr/bin/perl

################################################################################
# run_radmind.pl
#
# This script runs radmind and reboots.
#
# Requires the radmind client tools (and a running server obviously).
# (See http://rsug.itd.umich.edu/software/radmind)
#
# Can also be used with iHook
# (See http://rsug.itd.umich.edu/software/ihook)
#
# Copyright (c) 2002 University of Utah Student Computing Labs.
# All Rights Reserved.
#
# With some of Greg Neagle's modifications, Walt Disney Feature Animation
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appears in all copies and
# that both that copyright notice and this permission notice appear
# in supporting documentation, and that the name of The University
# of Utah not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission. This software is supplied as is without expressed or
# implied warranties of any kind.
#
#
# 2006.12.11
#    Added pre run tasks from a command file at /var/radmind/client/conf/radmind_prerun_tasks.conf
#
# 2006.12.09
#    Added postm config (config options derived from main command file /var/radmind/client/command.K, options so far include CASE_SENSITIVE)
#    Added case senstitive support
#
# 2006.09.28
#    Added n radmind server support
#
# 2006.05.08
#    Added Greg's progress bar stuff
#
# 2004_05_05
#    Totally redid this script, added lots of functions, gave it parameters
#    so it will act as prepForOverload, instead of having a totally different
#    script.
#
# 2004_01_30
#    Added logging to system commands and a few other tweaks
#    
#
################################################################################

use Getopt::Std;

use lib "/Library/Xhooks/Modules/";
require "xhooks/lib/config.pl";

my $run_maintenance_startup_trigger = "/Library/Preferences/Xhooks/triggerfiles/run_maintenance";
my $run_startup_post_maintenance = "/Library/Preferences/Xhooks/triggerfiles/run_startup_post_maintenance";
my $radmind_finished = "/Library/Preferences/Xhooks/triggerfiles/radmind_finished";

my $availablity_lib = "/Library/Xhooks/Modules/edu.utah.scl.availability/availability_lib.pl";
if ( -e $availablity_lib ) {
  require "edu.utah.scl.availability/availability_lib.pl";
}


################################################################################
# INSTRUCTIONS:
#
# Just run this script.  It will run radmind.  If you want to use this script
# with iHook, use a launcher script that contains this:
#
# $path_to_ihook = "/.../iHook.app/Contents/MacOS/iHook";
# $path_to_run_radmind_script = "/.../run_radmind.pl";
# system "$path_to_ihook --script=$path_to_run_radmind_script";
#
# To run it without iHook messages, put a "-n":
# 
# system "$path_to_run_radmind_script -n"
#
# To run it without rebooting or running post scripts, specify "-p" (prepForOverload)
# You probably want to specify "-n" to limit the printing too.
#
# system "$path_to_run_radmind_script -pn"
#
################################################################################


################################################################################
# Get args
#

getopts( "aiInp", \%options );

my $use_ihook;
my $automated_or_admin;
my $fsdiffpath;
my $lapply_options = ( defined $LAPPLY_OPTIONS ) ? "$LAPPLY_OPTIONS" : "";
my $ktcheck_options = ( defined $KTCHECK_OPTIONS ) ? "$KTCHECK_OPTIONS" : "";
my $fsdiff_options = ( defined $FSDIFF_OPTIONS ) ? "$FSDIFF_OPTIONS" : "";

if ( defined $options{i} ) {
  $use_ihook = 0;
} else {
  $use_ihook = 1;
}

if ( defined $options{'p'} ) {
  $automated_or_admin = "admin";
} else {
  $automated_or_admin = "automated";
}

if ( defined $options{'a'} ) {
  $fsdiffpath = "/";
} else {
  $fsdiffpath = ".";
}
if ( defined $FSDIFF_PATH ) {
  $fsdiffpath = $FSDIFF_PATH;
}

if ( ! ( defined $options{'n'} or ( defined $RADMIND_CHECKSUM and $RADMIND_CHECKSUM eq "False" ) ) ) {
  # This is the default behavior (no variable or no flag)
  $lapply_options .= " -c sha1";
  $fsdiff_options .= " -c sha1";
}

if ( defined $options{'I'} or ( defined $CASE_SENSITIVE and $CASE_SENSITIVE eq "False" ) ) {
  # case insenstive now!
  $lapply_options .= " -I"; 
  $fsdiff_options .= " -I";
}

if ( defined $RADMIND_PORT ) {
  $ktcheck_options .= " -p $RADMIND_PORT";
  $lapply_options .= " -p $RADMIND_PORT";
}

if ( defined $RADMIND_AUTH_LEVEL ) {
  $ktcheck_options .= " -w $RADMIND_AUTH_LEVEL";
  $lapply_options .= " -w $RADMIND_AUTH_LEVEL";
}

if ( defined $RADMIND_COMPRESSION_LEVEL ) {
  $ktcheck_options .= " -Z $RADMIND_COMPRESSION_LEVEL";
  $lapply_options .= " -Z $RADMIND_COMPRESSION_LEVEL";
}

if ( defined $RADMIND_FIX_HARD_LINK_SORT_ORDER and $RADMIND_FIX_HARD_LINK_SORT_ORDER = 1 ) {
  $fix_hard_link_sort_order = 1;
} else {
  $fix_hard_link_sort_order = 0;
}

################################################################################
# Stuff (kinda confusing)
#


# Where to run radmind from
$radmind_root_dir = "/"; # If the fsdiffpath is "." (see above) then you can
                         # change this if you want to radmind a volume like "/Volumes/Managed Disk/"
$run_radmind_command_file_header_lines = 8;

$iHookBackground1 = "/Library/Preferences/Xhooks/radmind/performing_maintenance1.gif"; # pre-ktcheck graphic
$iHookBackground2 = "/Library/Preferences/Xhooks/radmind/performing_maintenance2.gif"; # ktcheck graphic
$iHookBackground3 = "/Library/Preferences/Xhooks/radmind/performing_maintenance3.gif"; # fsdiff graphic
$iHookBackground4 = "/Library/Preferences/Xhooks/radmind/performing_maintenance4.gif"; # lapply graphic
$iHookBackground5 = "/Library/Preferences/Xhooks/radmind/performing_maintenance5.gif"; # post-lapply graphic

# Error catching:
$debug_overloads = 0;

# Logs
$errorlog = "/var/log/radmind_error.log";
$location_of_log = "/var/log/run_radmind_script_output";

# Command output files
$ktcheckoutput = "/var/log/ktcheck_output.log";
$fsdiffoutput = "/var/log/fsdiff_output.T";
$lapplyoutput = "/var/log/lapply_output.log";
$lapplyinput = "/var/log/lapply_input.T";
$ignoreList = "/Library/Preferences/Xhooks/radmind_fsdiff_ignore_list";

# The following number is the times this script will roll the output files
$ktcheckoutput_backup_number = 5;
$fsdiffoutput_backup_number = 5;
$lapplyoutput_backup_number = 5;
$errorlog_backup_number = 5;

# The commands
$ktcheck1 = "/usr/local/bin/ktcheck -c sha1 $ktcheck_options -h "; # between $ktcheck1 and $ktcheck2 is the
$ktcheck2 = " 2>> \"$errorlog\"";                             # radmind server, to be determined later....
$fsdiffCommand = "/usr/local/bin/fsdiff -A $fsdiff_options -o $fsdiffoutput -% $fsdiffpath 2>> \"$errorlog\"";
$lapply1 = "/usr/local/bin/lapply -F $lapply_options -h "; # between $lapply1 and $lapply2 is the
$lapply2 = " \"$lapplyinput\" 2>> \"$errorlog\"";                # radmind server, to be determined later....

$os_upgrade = 0; # dont' change this, this script auto detects.

# If radmind fails, it will run again (if there is a ktcheck update) this many times.
$max_tries = 3;

my $pause = 0;

################################################################################
# Run
#

if ( $#RADMIND_SERVERS < 0 ) {
  print "No radmind server set.  Quiting.\n";
  exit 1;
}

if ( $use_ihook ) {
  ################################################################################
  # Turn buffering off
  #
  $|++;
  $oldhandle = select ( STDERR );
  $|++;
  select ( $oldhandle );

  print "%UIMODE AUTOCRATIC\n";
  print "%BACKGROUND $iHookBackground1\n";
  print "%WINDOWZOOM DISABLE\n";
  print "%BECOMEKEY\n";
  print "%WINDOWLEVEL HIGH\n";
  #print "%BACKGROUNDSCALING TOFIT\n"; #TOFIT/PROPORTIONALLY/NONE
  #print "%WINDOWSIZE 640 480\n";
  #print "%CANCEL REMOVE\n";
  print "%0\n";
}

entman_logger( "window drawer", "Performing Maintenance" );
entman_logger( "drawer", `/bin/date` );

wait_for_network();

################################################################################
# Report that this computer is out of order
if ( defined $AVAILABILITY_URL and $AVAILABILITY_URL ne "" and -e $availablity_lib ) {
  report_availability ( "unavailable_maintenance", "radmind" );
}

################################################################################
# Keep track of time, write to log
#
$rightnow = time; # keep track of time
entman_logger( "log", "Started" );

unlink $radmind_finished; # remove the Radmind finished file so that we know RM is running.

################################################################################
# Roll logs
#
entman_logger( "drawer", "Rolling logs" );
roll_log( $ktcheckoutput, $ktcheckoutput_backup_number );
roll_log( $fsdiffoutput, $fsdiffoutput_backup_number );
roll_log( $lapplyoutput, $lapplyoutput_backup_number );
roll_log( $errorlog, $errorlog_backup_number );

################################################################################
# Run pre-radmind scripts (the RADMIND_PRERUN_COMMANDS array contains them)
#

if ( defined @RADMIND_PRERUN_COMMANDS and $#RADMIND_PRERUN_COMMANDS >= 0 )
{
  entman_logger( "log", "Shutting down services" );
  foreach $i ( @RADMIND_PRERUN_COMMANDS )
  {
    entman_logger( "log", "Executing: $i" );
    system "$i";
  }
}

$rserver = findRadmindServer ( @RADMIND_SERVERS );
if ( ! $rserver ) {
  system "/bin/echo \"No server can be found!\" >> \"$errorlog\" 2>&1";
  radmindFailed();
  finalcleanup( "failed" ); # this subroutine will reboot if this script is run automated
  exit 0;
}

preRadmindDeleteFiles();

################################################################################
# Delay so we don't hit the server at the same time... not sure this is causing
# the "snet_getline_multi: Operation timed out" messages, but it is worth a try.
if ( $automated_or_admin eq "automated" ) {
  sleep int(rand(10));
}

################################################################################
# Run the radmind tools
#
$radmind_result = run_radmind();

################################################################################
# Post radmind stuff
#

if ( $radmind_result == 0 ) {
  radmindSuccessful();
  finalcleanup( "successful" ); # this subroutine will reboot if this script is run automated
} else {
  system "/bin/echo radmind_result is greater than 0 >> \"$errorlog\"";
  radmindFailed();
  finalcleanup( "failed" ); # this subroutine will reboot if this script is run automated
}

exit 0;

sub uncomment_file_saveas {
  my ( $source, $dest ) = @_;
  if ( -e $source ) {
    if ( open INFILE, "< $source" ) {
      if ( open OUTFILE, "> $dest" ) {
  
        while ( <INFILE> ) {
          print OUTFILE substr( $_, 1 );
        }
        close ( INFILE );
        close ( OUTFILE );
      } else {
        entman_logger ( "Could not save $dest: $!" );
      }
    } else {
      entman_logger ( "Could not open $source: $!" );
    }
  }
}

################################################################################
sub run_radmind {

  chdir ( $radmind_root_dir );

  $current_try = 1;
  while ( 1 ) {

    ###########################
    # run ktcheck
    ###########################
    print "%BACKGROUND $iHookBackground2\n" if $use_ihook;
    $result = execute_ktcheck( $ktcheck1.$rserver.$ktcheck2, $ktcheckoutput, $current_try );
    return $result if $result != 0;

    # iHook directive.
    print "%20\n" if $use_ihook;

    # upgrade stuff (when changing from 10.2 to 10.3)
    do_upgrade_stuff();

    # Print command file info
    system "/usr/bin/head -n $run_radmind_command_file_header_lines /var/radmind/client/command.K";

    ################################################################################
    # Radmind Prerun Tasks
    # This code looks for a file named /var/radmind/client/conf/radmind_prerun_tasks.conf,
    # removes the first char from every line (should be a # char), and then executes it.
    my $radmind_prerun_tasks_source = "/var/radmind/client/conf/radmind_prerun_tasks.conf";
    my $radmind_prerun_tasks_dest = "/Library/Preferences/Xhooks/radmind_prerun_tasks.pl";
    uncomment_file_saveas($radmind_prerun_tasks_source, $radmind_prerun_tasks_dest);
    if ( -e $radmind_prerun_tasks_dest ) {
      entman_logger( "drawer", "Launching Radmind Prerun Tasks" );
      system "/usr/bin/perl", $radmind_prerun_tasks_dest;
    }

    ###########################
    # run fsdiff
    ###########################
    print "%BACKGROUND $iHookBackground3\n" if $use_ihook;
    # fsdiff this file once to get rid of the error that sometimes occurs.
    system "/usr/local/bin/fsdiff -c sha1 -1 /.hotfiles.btree";
    
    # Remove test triggerfile
    unlink "$entman_folder/system/delay_lapply";
    
    # Do the real fsdiff
    $result = execute_fsdiff( $fsdiffCommand, $fsdiffoutput );
    return $result if $result != 0;

    # iHook directive.
    print "%40\n" if $use_ihook;

    ################################################################################
    # filter output from fsdiff
    #
    if ( -e $ignoreList ) {
      system "/bin/cat \"$fsdiffoutput\" | /usr/bin/fgrep -v -f \"$ignoreList\" > \"$lapplyinput\"";
    } else {
      system "/bin/cp \"$fsdiffoutput\" \"$lapplyinput\"";
    }

    ###########################
    # process fsdiff output
    ###########################

    # Fix hardlinks created when using converted case sensitive transcripts!
    if ( $fix_hard_link_sort_order ) {
       #system q{cat "${trans}" | awk '{ if ( $1 == "h" ) { if ( tolower($2) < tolower($3) ) { if ( system( "test -e " $3 ) != 0 ) { system( "touch " $3 ); } } } }'};
    }

    ###########################
    # run lapply
    ###########################
    print "%BACKGROUND $iHookBackground4\n" if $use_ihook;
    $result = execute_lapply( $lapply1.$rserver.$lapply2, $lapplyoutput, $lapplyinput );

    ###########################
    # Play again?
    ###########################
    if ( $result == 0 ) {
      if ( -e $errorlog ) {
        my $errorlog_size = ( stat ( $errorlog ) )[7];
        if ( $errorlog_size > 0 ) {
          # Error, but only if $debug is on (in other words, radmind "finished",
          # but this branch should only happen if lapply errors were ignored)
          return 1; # Failed
        } else {
          return 0; # Successful
        }
      } else {
        return 0; # Successful
      }
    } else {
      if ( $result == 1 ) {  
          $system_modified_message = "system was modified";
      } else {
          $system_modified_message = "system was not modified";
      }
      if ( $current_try >= $max_tries ) {
        entman_logger( "window log drawer", "Ouch, lapply error, $system_modified_message.  See log for errors." );
        return 1; # Failed
      } else {
        entman_logger( "window log drawer", "Ouch, lapply error, $system_modified_message. Trying again." );
      }
    }
    $current_try++;

  }
}

##################################################################################
# Radmind Loadballancing
#
sub findRadmindServer {
  my ( @server_list ) = @_;

  entman_logger( "drawer", "Finding which radmind server to use and checking to see if it is reachable" );

  my $counter = 0;
  my @new_server_list = ();
  while ( $#new_server_list < 0 ) {
    if ( $counter > 60 ) {
      entman_logger( "log drawer window", "No radmind servers found!" );
      return;
    }
    @new_server_list = findUpServers ( 6662, @server_list );
    sleep 1;
    $counter++;
  }

  if ( ! -e $use_shadow_server or $automated_or_admin eq "admin" ) {

    $rserver = @new_server_list[0];
    entman_logger( "drawer", "No load balancing: Using $rserver because it is the first server that was up." );

  } else {

    my $my_ip = get_ip();
    my @ip_parts = split ( '\.', $my_ip );
    # Send this computer to one of the servers
    $rserver = $new_server_list[ $ip_parts[3] % ( $#new_server_list + 1 ) ];
    
    # This wasn't as funny as we thought, so it is comment out for now, until we can figure out how to make it funnier...
    #if ( $ip_parts[3] % 3 == 0) {
    #  system "say oink oink oink";
    #} elsif ( $ip_parts[3] % 3 == 1) {
    #  system "say moo moo moo";
    #} else {
    #  system "say quack quack quack";
    #}
    
    entman_logger( "drawer", "Load balancing: I'm using $rserver, server number ".($#new_server_list+1).", my ip is $ip_parts[3]." );

  }

  return $rserver;

}

################################################################################
# Save time by deleting w/ rm lots of known files that need to be deleted!
#
sub preRadmindDeleteFiles {

  ################################################################################
  # Delete home folders
  #

  if ( -e $homeFolderBackups ) {
    entman_logger( "drawer", "Deleting $homeFolderBackups" );
    system "/usr/bin/chflags -R nouchg \"$homeFolderBackups\" >> $location_of_log 2>&1";
    system "/bin/rm -rf \"$homeFolderBackups/\"* >> $location_of_log 2>&1";
  }

  if ( defined $AUTOLOGIN_HOME_DIR and -e $AUTOLOGIN_HOME_DIR and defined $RADMIND_DELETE_AUTOLOGIN_HOME_DIR and $RADMIND_DELETE_AUTOLOGIN_HOME_DIR eq "True" ) {
    entman_logger( "drawer", "Deleting $AUTOLOGIN_HOME_DIR" );
    system "/usr/bin/chflags -R nouchg \"$AUTOLOGIN_HOME_DIR\" >> $location_of_log 2>&1";
    system "/bin/rm -rf \"$AUTOLOGIN_HOME_DIR\" >> $location_of_log 2>&1";
  }
  
  # Right here... move the folder to lost and found instead.
  if ( defined $AUTHLOGIN_HOME_DIR and -e $AUTHLOGIN_HOME_DIR and defined $RADMIND_DELETE_AUTHLOGIN_HOME_DIR and $RADMIND_DELETE_AUTHLOGIN_HOME_DIR eq "True" ) {
    entman_logger( "drawer", "Deleting $AUTHLOGIN_HOME_DIR" );
    system "/usr/bin/chflags -R nouchg \"$AUTHLOGIN_HOME_DIR\" >> $location_of_log 2>&1";
    system "/bin/rm -rf \"$AUTHLOGIN_HOME_DIR\" >> $location_of_log 2>&1";
  }

  # This section could this hang the system if there is HD corruption?  It happens...
  if ( -e $global_home_folder_lost_and_found and $radmind_delete_home_folder_cache_switch ) {
    entman_logger( "drawer", "Deleting $global_home_folder_lost_and_found" );
    system "/usr/bin/chflags -R nouchg \"$global_home_folder_lost_and_found\" >> $location_of_log 2>&1";
    system "/bin/rm -rf \"$global_home_folder_lost_and_found/\"* >> $location_of_log 2>&1";
  }

  ################################################################################
  # Delete stuff ignored by radmind, but still need to be cleaned up.
  #

  if ( defined $RADMIND_DELETE_CACHES and $RADMIND_DELETE_CACHES eq "True" ) {

    #/Library/Caches/
    #
    if ( $backup_launch_services_cache ) {
      system "/bin/cp /Library/Caches/com.apple.LaunchServices-0140.csstore /var/tmp/";
      system "/bin/cp /Library/Caches/com.apple.LaunchServices.6B.csstore /var/tmp/";
    }

    entman_logger( "drawer", "Deleting /Library/Caches/" );
    system "/usr/bin/chflags -R nouchg /Library/Caches/* >> $location_of_log 2>&1";
    system "/bin/rm -rf /Library/Caches/* >> $location_of_log 2>&1";
    system "/bin/rm -rf /Library/Caches/.* >> $location_of_log 2>&1";

    if ( $backup_launch_services_cache ) {
      system "/bin/mv /var/tmp/com.apple.LaunchServices-0140.csstore /Library/Caches/";
      system "/bin/mv /var/tmp/com.apple.LaunchServices.6B.csstore /Library/Caches/";
    }

    #/System/Library/Caches/
    #
    entman_logger( "drawer", "Deleting /System/Library/Caches/" );
    system "/usr/bin/chflags -R nouchg /System/Library/Caches/* >> $location_of_log 2>&1";
    system "/bin/rm -rf /System/Library/Caches/* >> $location_of_log 2>&1";
    system "/bin/rm -rf /System/Library/Caches/.* >> $location_of_log 2>&1";

    # /var/tmp/
    #
    # delete everything not owned by root or nobody (dangerous...?)
    entman_logger( "drawer", "Cleaning /var/tmp" );
    system "find /var/tmp \! -user 0 -and \! -user -2 -delete";
#    @delete_these = `/usr/bin/grep -l PPD-Adobe /var/tmp/* 2> /dev/null`;
#    foreach $delete_me ( @delete_these ) {
#      chomp $delete_me;
#      system "/bin/rm $delete_me";
#    }

    #/Library/Logs/Console - Remove too old
    system "find /Library/Logs/Console/ -maxdepth 1 -mtime +30 -delete >> $location_of_log 2>&1";
    system "find /Library/Logs/Console/ -mtime +30 -delete >> $location_of_log 2>&1";

    #/Library/Logs/Console - Remove too big (I just picked a number that seemed good...)
    my @usageList = `du -kd0 /Library/Logs/Console/*`;
    foreach $ii ( @usageList ) {
      my ( $size, $path ) = split ' ', $ii;
      if ( $size gt "3000" ) {
        if ( $path =~ /"/ ) {
          entman_logger( "drawer", "ERROR! Possible script injection in /Library/Logs/Console" );
        } else {
          system "rm -fr -- \"$path\"";
        }
      }
    }

  }

}

################################################################################
# This function executes ktcheck.
# Runs ktcheck.  If $current_try is 1, it will return 0 (success) if there were
# updates or no updates.  If it is greater than 1, it will return success only when
# there are updates.  This is so that if there is an error on the first time,
# it will try again if there are updates (meaning the update probably caused the
# error).
#
sub execute_ktcheck {
  my ( $command, $ktcheckoutput, $current_try ) = @_;
  entman_logger( "drawer window", "Checking for Updates" );
  $result = execute_command ( $command, $ktcheckoutput );
  if ( $result == 0 ) {
    entman_logger( "drawer", "No updates" );
    if ( $current_try > 1 ) {
      #There are no updates.  Quit.
      entman_logger( "window log drawer", "There are no updates, and updates are required to continue.  Quiting" );
      return 1;
    }
  } elsif ( $result == 1 ) {
    entman_logger( "drawer", "Updates found" );
  } else {
    entman_logger( "window log drawer", "ktcheck encountered a fatal error: $result" );
    return 1;
  }
  return 0;
}

################################################################################
# This function executes fsdiff.
# Just runs fsdiff and returns error if there is one.
#
sub execute_fsdiff {
  my ( $fsdiffCommand, $fsdiffoutput ) = @_;
  entman_logger( "drawer window", "Scanning File System" );
  #$result = execute_command ( "$fsdiffCommand", $fsdiffoutput );
  $result = system "$fsdiffCommand";

  if ( $result != 0 ) {
    entman_logger( "window log drawer", "fsdiff encountered a fatal error: $result" );
    return 1;
  }
  return 0;
}

################################################################################
# This function executes lapply.
# Runs lapply.  Returns -1 if there was an error but it thinks it should try
# again.  Returns 0 when all is well.  Returns 1 when it gets an error and
# thinks it should quit.
#
sub execute_lapply {
  my ( $command, $lapplyoutput, $lapplyinput ) = @_;
  entman_logger( "drawer window", "Updating File System" );
  $fsdsize = ( stat ( $lapplyinput ) )[7];
  if ( $fsdsize > 0 ) {
    $result = execute_gregs_lapply( $command, $lapplyoutput );
    return $result;
  } else {
    entman_logger( "window log drawer", "fsdiff found no changes, which is more than unlikely..." );
    return 0;
  }
}

################################################################################
# This subroutine executes lapply  and saves the output to a log,
# prints it to stdout, and saves stderr to a log as well.
# It also provides percent-done feedback for use by iHook.
# Greg's subroutine slightly modified by James Reynolds :)
#
sub execute_gregs_lapply {
  my ( $thecommand, $path_to_log ) = @_;
  my $lapply_line_count = 0;
  my $fsdiff_line_count = int ((split ' ', `/usr/bin/wc -l \"$lapplyinput\"`)[0]);
  my $percentDone;

  open ( COMMAND, "$thecommand |" );
  open ( LOG, ">>$path_to_log" );
  while ( <COMMAND> ) { 

    print STDERR;
    print LOG;

    if ( $lapply_line_count % 10 == 0 ) {
      $percentDone = int ($lapply_line_count/$fsdiff_line_count*99);
      $percentDone = 100 if $percentDone > 100;
      print "%$percentDone\n";
    }
    $lapply_line_count++;

  }
  close ( LOG );
  close ( COMMAND );
  return $? >> 8;
}

################################################################################
# This subroutine executes the radmind commands and saves the output to a log,
# prints it to stdout, and saves stderr to a log as well.
#
sub execute_command {
  my ( $thecommand, $path_to_log ) = @_;
  open ( COMMAND, "$thecommand |" );
  open ( LOG, ">>$path_to_log" );
  while ( <COMMAND> ) {
    print STDERR;
    print LOG;
  }
  close ( LOG );
  close ( COMMAND );
  return $? >> 8;
}

################################################################################
# If radmind was successful, do these things
#
sub radmindSuccessful {
  print "%BACKGROUND $iHookBackground5\n" if $use_ihook;

  # No error
  unlink $errorlog;
  system "/bin/date > $maintenance_lastrun";

  entman_logger( "drawer window", "Radmind finished, doing other tasks" );

  # iHook directive.
  print "%80\n" if $use_ihook;
}

################################################################################
# If radmind died, this is the cleanup subroutine.
#
sub radmindFailed {
  # print error log to stdout
  print "%BACKGROUND $iHookBackground5\n" if $use_ihook;
  system "/bin/echo Server used: $rserver >> \"$errorlog\"" if ( $rserver );
  system "/bin/echo 'Error generated on' >> \"$errorlog\"";
  system "/bin/date >> \"$errorlog\"";
  open ( LOGGY, "<$errorlog" );
  while ( <LOGGY> ) {
    print;
  }
  close ( LOGGY );

  system "/bin/rm $run_maintenance_startup_trigger >> $location_of_log 2>&1"; # remove this file or it will keep rebooting!
}

################################################################################
# Final cleanup before rebooting.
#
sub finalcleanup {
  my ( $status ) = @_;

  unlink "/System/Library/Extensions.mkext";
  unlink "/System/Library/Extensions.kextcache";

  if ( $os_upgrade ) {

    print "%100\n";

    entman_logger( "window", "OS upgraded, force reboot me now if needed" );

    exit_script(); # do not detach here or else it will go to login panel in many cases...

  } elsif ( $automated_or_admin eq "automated" ) {

    # post maintenance trigger file
    system "touch \"$run_startup_post_maintenance\"";
    
    # How long did script take?
    $rightnowest = time;
    $totaltime = $rightnowest - $rightnow;
    $remote_logger_message = "Radmind $status, total script time is $totaltime seconds";

    system "\"$remote_logger_script\" Radmind \"$remote_logger_message\" >> $location_of_log 2>&1 &";
    sleep 15; # give the above 15 seconds.
    entman_logger( "log", "$remote_logger_message" );

    print "%100\n" if $use_ihook;
  
    system "touch $radmind_finished";
  
    #if ( $checkForRestart ) {
    entman_logger( "window log drawer", "Restarting..." );
    sleep 5;
    exit_script(); # do not detach here or else it will go to login panel in many cases...

  } else {

    # post maintenance
    print "If you want this to behave just like the automated scripts, execute this commmand:\n";
    print "touch \"$run_startup_post_maintenance\"\n";

    system 'osascript -e "say \"Attention. I am now ready to be modifed in preperation for creating an overload.\""';

    # touch trigger file or radmind will keep running.
    system "touch $radmind_finished";

  }
}

################################################################################
# This subroutine has not been tested
#
sub checkForRestart {
  # If the user who logged in is "radmind", this script should reboot no matter what.

  #$exit_value = system "fgrep -qf $rebootList $fsdiffoutput";
  #if ( $exit_value == 0 ) {
  #  restart = 1;
  #} 
}

sub encode_path {
  my ( $path ) = @_;
  $path =~ s/ /\\b/g;; # convert spaces
#  my $path =~ s/\t/\\t/g;; # convert tabs
#  my $path =~ s/\n/\\n/g;; # convert newline
#  my $path =~ s/\r/\\r/g;; # convert return
#  my $path =~ s/\/\\/g;; # convert backslashes
  return $path;
}

sub decode_path {
  my ( $path ) = @_;
  $path =~ s/\\b/ /g;;
#  my $path =~ s/\\t/\t/g;; # convert tabs
#  my $path =~ s/\\n/\n/g;; # convert newline
#  my $path =~ s/\\r/\r/g;; # convert return
#  my $path =~ s/\\/\/g;; # convert backslashes
  return $path;
}

sub get_path_of_transcript_line {
  my ( $transcript_line, $folder_only ) = ( @_ );
  my @parts = split ' ', $input;
  my $path = "";
  if ( $#parts >= 4 ) {
    if ( $parts[0] eq "d" ) {
      $path = $parts[1];
    } elsif ( $parts[1] eq "d" ) {
      $path = $parts[2];
    } else {
      if ( $parts[0] eq "f" || $parts[0] eq "a" ) {
        $path = $parts[1];
      } elsif ( $parts[1] eq "f" || $parts[1] eq "a" ) {
        $path = $parts[2];
      } else {
        return ""; # exit if it isn't a file or apple file.
      }
      if ( defined $folder_only and $folder_only ) {
        remove_file_from_path( $path );
      }
    }
    $path = decode_path( $path );
  }
  return $path;
}

sub remove_file_from_path {
  my ( $path ) = @_;
  # remove the file off the end of the line (kinda a dumb way to do it, but oh well...)
  my @folders = split '/', $path;
  my $depth = @folders;
  $path = "";
  for ( $i = 0 ; $i < $depth - 1 ; $i++ ) {
    $path .= $folders[$i]. "/";
  }
  if ( $path ne "./" ) {
    chop $path;
  }
  return $path;
}

################################################################################
# Do this stuff if I detect an os change (like from 10.2 to 10.3)
# For this to work, you must have the version number listed in the base overload
# AND your base overload must have "os_base" in it.
sub do_upgrade_stuff {
  $booted_os_version = `sw_vers | grep ProductVersion`;
  $command_file_version = `grep "# OS" /var/radmind/client/command.K`;
  if ( $command_file_version ne "" and $booted_os_version ne "" ) {
    $booted_os_version = substr $booted_os_version, ( index $booted_os_version, "10." ), 4;
    $command_file_version = substr $command_file_version, ( index $command_file_version, "10." ), 4;
    if ( $booted_os_version ne $command_file_version ) {
      $os_upgrade = 1;
      entman_logger( "drawer", "WARNING! OS upgrade in progress!!! Force reboot may be necessary" );
      if ( $booted_os_version eq "10.2" ) {
        system "/usr/bin/killall cron >> $location_of_log 2>&1";
        system "/usr/bin/killall cupsd >> $location_of_log 2>&1";
        kill_process( "ARD Helper", "root", "" );
        kill_process( "ARD Agent", "root", "" );
        kill_process( "perl", "root", "idlescript.pl" );
      } elsif ( $booted_os_version eq "10.3" ) {
        system "rm -Rf /Library/Printers/hp/Utilities";
      }     
    }
  } else {
    entman_logger( "drawer log", "Could not find OS versions" );
  }
}

################################################################################
sub kill_process  {
  my ( $appToCheck, $username, $message_to_grep_for ) = @_;
  $message_to_grep_for =~ s/"//g;
  open ( PROC, "/bin/ps -wwxo pid,command -U $username |" );   
  @proclist = <PROC>;
  close PROC;
  @the_list = grep ( /$appToCheck/i, @proclist );
  @kill_me = grep ( /$message_to_grep_for/i, @the_list );
  if ( defined @kill_me ) {
    @pid = split( ' ', $kill_me[0] );
    if ( $pid[0] ne "" ) {
      system "/bin/kill $pid[0] >> $location_of_log 2>&1";
      return 1;
    }
  }
  return 0;
}

sub exit_script {
  # put your prefered exit method here.
  if ( $pause ) { print "You should reboot now\n"; exit; }
  system "/sbin/reboot";
  exit;
}
