#!/usr/bin/perl

################################################################################
# rsync_to_shadow.pl
#
# This script rsyncs a master radmind server to a slave
#
# Copyright (c) 2002 University of Utah Student Computing Labs.
# All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appears in all copies and
# that both that copyright notice and this permission notice appear
# in supporting documentation, and that the name of The University
# of Utah not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission. This software is supplied as is without expressed or
# implied warranties of any kind.
#
# 2005.12.30 - Now keeps a catalog of overloads and their checksum. This
#              catalog is synced to the other server.  So the other server
#              can read the file and delete overloads not listed in it!
#              This script will also only sync overloads with changed
#              transcripts to make it much faster (hopefully!)
# 2005.??.?? - Changed it so it treats /var/radmind/transcript the same as 
#              /var/radmind/file
# 2005.01.05 - Changed how /var/radmind/file was handled.  Now I can tell this 
#              script to scan the entire file directory or only each folder in
#              the file directory.
#
################################################################################

################################################################################
# Change these
#
$source_base = "/var/radmind";
$destination_base = "radmind\@10.0.1.2:/var/radmind";
$logger = "/usr/bin/logger -i -t $0"; # Your logger
$noisy_log = 0;
$rsync_output = "/var/log/rsync_output"; # this should be a folder
if ( ! -d $rsync_output ) {
  if ( -e $rsync_output ) {
    unlink $rsync_output;
  }
  mkdir $rsync_output;
}

$rsync_output_backup_number = 5;

@rsync_these_folders = ( "command", "config", "special", "overload_catalog" ); 
# Do not include "file" or "transcript" in the above list, because it is handled with 
# the $scan_file_folder_seperately variable.

$scan_file_folder_seperately = 1;
if ( defined $ARGV[0] ) {
  $scan_file_folder_seperately = 0;
}

################################################################################
# Internal stuff
#

$options = "--rsh=ssh --hard-links --progress --delete";                 
$cl_to_gui_messenger = "/Applications/CL_to_GUI_Messenger.app/Contents/MacOS/CL_to_GUI_Messenger";
$rsync_command = "/usr/bin/rsync -vrl $options ";   

# not used anymore:
#$rsync_command_non_recursive = "/usr/bin/rsync -vl $options "; # remove recursive

$last_source = "bla"; # don't change this
$output_file = "$rsync_output/rsync.out";
$error_file = "$rsync_output/rsync.err";
$error_bool = 0;
$last_error_message = "";

################################################################################
# Log
#

system "$logger \"Started.\""; 

if ( ! -d $rsync_output ) {
  system "$logger \"NO valid log location, exiting.\""; 
  exit 1;
}

# Handle the log file
roll_log ( "$output_file", "$rsync_output_backup_number" );
roll_log ( "$error_file", "$rsync_output_backup_number" );

################################################################################
# Get list of folders to sync
#

@tempy = ();
if ( $scan_file_folder_seperately ) {

  my $catalog_file = "$source_base/overload_catalog";
  my $folder_to_catalog = "$source_base/transcript";
  unlink "$source_base/transcript/.DS_Store";

  # Read old overload catalog
  my %old_overload_catalog = ();
  if ( -e $catalog_file ) {
    open ( FILE, "< $catalog_file" ) or die "Can't read from $catalog_file : $!";
    while( <FILE> ) {
      my ( $o_path, $o_cksum ) = ( split /\s+/, $_ );
      $old_overload_catalog{ $o_path } = $o_cksum;
    }
    close FILE;
  }

  # remove catalog
  unlink $catalog_file;

  # scan all overloads
  @transcripts = getFileFolderList ( $folder_to_catalog );
  my @new_overload_catalog = ();
  foreach $transcript ( @transcripts ) {
    
    my $fsdiff = `/usr/local/bin/fsdiff -c sha1 -1 \"$folder_to_catalog/$transcript\"`;
    my ( $n_type, $n_path, $n_perm, $n_owner, $n_group, $n_mtime, $n_size, $n_cksum ) = ( split /\s+/, $fsdiff );

    # build new overload catalog
    open ( FILE, ">> $catalog_file" ) or die "Can't write to $catalog_file : $!";
    print FILE $n_path ." ". $n_cksum ."\n";
    close FILE;

    # build list of updated overloads
    if ( not ( defined $old_overload_catalog {$n_path} and $n_cksum eq $old_overload_catalog {$n_path} ) ) {
      push @tempy, "transcript/$transcript";
      if ( -d "$source_base/file/$transcript" ) {
        push @tempy, "file/$transcript";
      }
    }
  }
} else {
  push @tempy, "file";
  push @tempy, "transcript";
}

@rsync_these_folders = ( @tempy, @rsync_these_folders );

open ( LOG, ">>$output_file" );
print LOG `date`;
print LOG "Rsync starting.  Here is what I am going to sync:\n";
foreach $i ( @rsync_these_folders ) {
  print LOG "$source_base/$i\n";
}
close ( LOG );

################################################################################
# Run rsync
#
foreach $i ( @rsync_these_folders ) {
  run_rsync ( $rsync_command, "$source_base/$i", "$destination_base/$i" );
}

################################################################################
# Clean up
#

# Erase last message
if ( -e $cl_to_gui_messenger ) {
  if ( $last_source ne "bla" ) {
    killCLGUI ( "$cl_to_gui_messenger", "root", "$last_source" );
  }
}

# Display message if there was an error
if ( -e $cl_to_gui_messenger ) {
  if ( $error_bool ) {
    system "\"$cl_to_gui_messenger\" \"There was an rsync error.\" \"$last_error_message\" &";
  }
}

system "$logger \"Finished.\"";

################################################################################
sub run_rsync {
  my ( $rsync_command, $source, $destination ) = @_;

  # Do GUI message
  if ( -e $cl_to_gui_messenger ) {
    if ( $last_source ne "bla" ) {
      killCLGUI ( "$cl_to_gui_messenger", "root", "$last_source" );
    }
    $last_source = $source;
    system "\"$cl_to_gui_messenger\" \"Running Rsync, please do not change any radmind stuff!\" \"$source\" &";
  }

  # check the to see if it is a directory and append "/" if it is...
  if ( -d $source ) {
    $source .= "/";
    $destination .= "/";
  }

  # Print log info
  open ( LOG, ">>$output_file" );
  print LOG `date`;
  print LOG "$rsync_command $source $destination\n\n";
  #print "$rsync_command $source $destination\n\n";

  # Run Rsync
  if ( $noisy_log ) {
    open ( COMMAND, "$rsync_command $source $destination 2>> $error_file |" );
    while ( <COMMAND> ) {
      print LOG;
    }
    close ( COMMAND );
  } else {
    system "$rsync_command $source $destination 2>> $error_file";  
  }

  # Check for errors
  $result = $? >> 8;
  if ( $result > 0 ) {
    $error_bool = 1;
    system "$logger \"Error: $source\"";
    $last_error_message .= $source;
    print LOG "There was an error.";
  }
  close ( LOG );

}

################################################################################
sub getFileFolderList {
  my ( $folder ) = @_;
  @dirlist=();
  if ( -d $folder ) {
    opendir ( DIR, "$folder" );
    @dirlist=readdir ( DIR );
    closedir ( DIR );
    @dirlist = splice ( @dirlist, 2 );
  } 
  return @dirlist;
}

################################################################################
# This script renames $path_to_log so that the lowest number is always the
# newest.  The oldest that is greater than $number_of_backups is deleted.
#  
sub roll_log {
  my ( $path_to_log, $number_of_backups ) = @_;
  if ( -e $path_to_log ) {
    if ( -e $path_to_log.".bak$number_of_backups" ) {    
      unlink $path_to_log.".bak$number_of_backups";
    }
    for ( $i = $number_of_backups ; $i > 1 ; $i-- ) {
      if ( -f $path_to_log.".bak".($i-1)) {
        system "/bin/mv -f \"$path_to_log".".bak".($i-1)."\" \"$path_to_log".".bak$i\"";
      }
    }
    system "/bin/mv -f \"$path_to_log\" \"$path_to_log".".bak1\"\n";
  } 
}

################################################################################
sub killCLGUI  {
  my ( $appToCheck, $username, $message_to_grep_for ) = @_;
  open ( PROC, "ps -wwxo pid,command -U $username |" );   
  @proclist = <PROC>;
  close PROC;
  @cl_to_gui_list = grep ( /$appToCheck/i, @proclist );
  @kill_me = grep ( /$message_to_grep_for/i, @cl_to_gui_list );
  @pid = split ( ' ', $kill_me[0] );
  if ( $pid[0] ne "" ) {
    system "kill $pid[0]";
  }
}
